#![cfg(test)]

use http::header::CONTENT_TYPE;
use oauth2::Scope;
use std::str::FromStr;
use time::{Duration, OffsetDateTime};
use url::Url;

use crate::{
    http::CONTENT_TYPE_JSON,
    standards::indieauth::{ensure_discovery, find_metadata, Scopes, VerifiedToken},
};

#[test]
fn scopes_serde() {
    assert_eq!(
        Ok(Scopes(vec![
            Scope::new("read".into()),
            Scope::new("profile".into())
        ])),
        serde_json::from_str(r#""read profile""#).map_err(|e| e.to_string())
    );

    assert_eq!(
        Ok(Scopes(vec![
            Scope::new("read".into()),
            Scope::new("profile".into())
        ])),
        serde_json::from_str(r#"["read", "profile"]"#).map_err(|e| e.to_string())
    );

    #[derive(serde::Deserialize, Debug, PartialEq, Eq)]
    struct K {
        scope: Scopes,
    }

    assert_eq!(
        Ok(K {
            scope: Scopes(vec![
                Scope::new("read".into()),
                Scope::new("profile".into())
            ])
        }),
        serde_qs::from_str("scope=read+profile").map_err(|e| e.to_string())
    );

    assert_eq!(
        Ok(K {
            scope: Scopes(vec![
                Scope::new("read".into()),
                Scope::new("profile".into())
            ])
        }),
        serde_qs::from_str("scope[]=read&scope[]=profile").map_err(|e| e.to_string())
    );
}

#[test]
fn verified_token() {
    let issued_at = OffsetDateTime::now_utc() - Duration::minutes(10);
    let expires_at = OffsetDateTime::now_utc() + Duration::days(8);
    assert_eq!(
        Some(serde_json::json!({
            "active": true,
            "me": "http://me.com/",
            "client_id": "http://client.id/",
            "scope": "read profile",
            "exp": expires_at.unix_timestamp(),
            "iat": issued_at.unix_timestamp()
        })),
        serde_json::to_value(VerifiedToken {
            active: true,
            me: "http://me.com".parse().unwrap(),
            client_id: "http://client.id".parse().unwrap(),
            scopes: Scopes::from_str("read profile").unwrap(),
            expires_at,
            issued_at
        })
        .ok()
    );

    assert_eq!(
        Some(serde_json::json!(VerifiedToken {
            active: true,
            me: "http://me.com".parse().unwrap(),
            client_id: "http://client.id".parse().unwrap(),
            scopes: Scopes::from_str("read profile").unwrap(),
            expires_at,
            issued_at
        })),
        serde_json::from_value(serde_json::json!({
            "active": true,
            "me": "http://me.com/",
            "client_id": "http://client.id/",
            "scope": "read profile",
            "exp": expires_at.unix_timestamp(),
            "iat": issued_at.unix_timestamp()
        }))
        .ok()
    );
}
