use crate::{
    http::CONTENT_TYPE_JSON,
    standards::micropub::{
        paging,
        query::{
            CategoryResponse, ConfigurationResponse, MatchingPropertyValuesMap, Query, Response,
            SourceQuery, SourceResponse,
        },
    },
};
use http::header::CONTENT_TYPE;
use microformats::types::{Class, KnownClass};
use oauth2::AccessToken;

use crate::{
    algorithms::ptd::Type,
    standards::micropub::{query::QueryKind, Error},
};

#[test]
#[cfg(feature = "experimental_channels")]
fn query_response_for_channels() {
    use super::extension;

    assert_eq!(
        Some(Response::Channel(extension::channel::QueryResponse {
            channels: vec![extension::channel::Form::Expanded {
                uid: "magic".to_string(),
                name: "Magic".to_string(),
                properties: serde_json::json!({"grr": "bark"}).try_into().unwrap()
            }],
            paging: Default::default()
        })),
        serde_json::from_value(serde_json::json!({
            "channels": [{
                "uid": "magic",
                "name": "Magic",
                "grr": "bark"
            }]
        }))
        .ok()
    )
}

#[test]
#[cfg(all(feature = "experimental_channels", feature = "experimental_syndication"))]
fn query_response_for_configuration() {
    assert_eq!(
        Ok(Response::Configuration(ConfigurationResponse {
            q: vec!["channels".to_owned()],
            category: vec!["tag".into()],
            media_endpoint: None,
            post_types: vec![Type::Note],
            channels: Default::default(),
            syndicate_to: Default::default()
        })),
        serde_json::from_value(serde_json::json!({
            "q": ["channels"],
            "post-types": ["note"],
            "category": ["tag"]
        }))
        .map(Response::Configuration)
        .map_err(crate::Error::JSON)
    );
}

#[test]
#[cfg(all(feature = "experimental_channels", feature = "experimental_syndication"))]
fn query_response_for_configuration_with_syndication() {
    use crate::standards::micropub::extension;

    assert_eq!(
        Ok(Response::Configuration(ConfigurationResponse {
            q: vec!["channels".to_owned()],
            category: vec!["tag".into()],
            media_endpoint: None,
            post_types: vec![Type::Note],
            channels: Default::default(),
            syndicate_to: vec![extension::syndication::Target {
                uid: "magic".into(),
                name: "cookie".into(),
                ..Default::default()
            }]
        })),
        serde_json::from_value(serde_json::json!({
            "q": ["channels"],
            "post-types": ["note"],
            "category": ["tag"],
            "syndicate-to": [
                {
                    "uid": "magic",
                    "name": "cookie"
                }
            ]
        }))
        .map(Response::Configuration)
        .map_err(crate::Error::JSON)
    );
}

#[test]
#[cfg(all(feature = "experimental_channels", feature = "experimental_syndication"))]
fn query_response_for_configuration_with_channels() {
    assert_eq!(
        Ok(Response::Configuration(ConfigurationResponse {
            q: vec!["channels".to_owned()],
            category: vec!["tag".into()],
            media_endpoint: None,
            post_types: vec![Type::Note],
            channels: Default::default(),
            syndicate_to: Default::default()
        })),
        serde_json::from_value(serde_json::json!({
            "q": ["channels"],
            "post-types": ["note"],
            "category": ["tag"]
        }))
        .map(Response::Configuration)
        .map_err(crate::Error::JSON)
    );
}

#[test]
fn query_response_for_categories() {
    assert_eq!(
        Ok(Response::Category(CategoryResponse {
            categories: vec!["jump".into(), "kick".into(), "spin".into()],
            pagination: Default::default()
        })),
        serde_json::from_value(serde_json::json!({
            "categories": ["jump", "kick", "spin"]
        }))
        .map_err(|e| format!("{:#?}", e)),
        "works without paging"
    );
}

#[test]
fn query_response_for_categories_with_paging() {
    use super::extension::Order;

    assert_eq!(
        Ok(CategoryResponse {
            categories: vec!["jump".into(), "kick".into(), "spin".into()],
            pagination: paging::Fields {
                paging: paging::Query {
                    order: Some(Order::Descending),
                    ..Default::default()
                },
                ..Default::default()
            }
        }),
        serde_json::from_value(serde_json::json!({
            "categories": ["jump", "kick", "spin"], "paging": { "order": "desc" }
        }))
        .map_err(|e| format!("{:#?}", e)),
        "works with paging"
    )
}
